"""
| Utilities for mine_init
"""
import os
import re
import subprocess
import email
import logging
from logging.handlers import BufferingHandler
from smtplib import SMTP, SMTPException


class MakeFileHandler(logging.FileHandler):
    """
    | A file handler class that ensures the logging dir is precreated
    """
    def __init__(self, filename, mode='a', encoding=None, delay=0):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        logging.FileHandler.__init__(self, filename, mode, encoding, delay)


class BufferingSMTPHandler(BufferingHandler):
    """
    | A log handler that buffers messages into memory to deliver them as a single email.
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, mail_server, mail_port, src_email, dest_email, capacity):
        BufferingHandler.__init__(self, capacity)
        self.mail_server = mail_server
        self.mail_port = mail_port
        self.src_email = src_email
        self.dest_email = dest_email
        self.subject = None
        self.description = None
        self.cc_list = None
        self.send = False

    def flush(self):
        if self.buffer and self.send:
            try:
                message = email.message.EmailMessage()
                message['Sender'] = self.src_email
                message['To'] = self.dest_email
                if self.cc_list:
                    message['Cc'] = ','.join(self.cc_list)
                message['Subject'] = self.subject
                body = self.description + '\r\n'
                for record in self.buffer:
                    log_line = self.format(record)
                    body = body + log_line + "\r\n"
                message.set_content(body)
                with SMTP(self.mail_server, self.mail_port) as smtp:
                    smtp.send_message(message)
            except SMTPException as error:
                self.handleError(error)
            self.buffer = []

    def send_email(self, subject, description, cc_list=None):
        """
        | Send email by calling flush after setting some last minute details.
        | :param subject: `string` Subject for outgoing email
        | :param description: `string` Opening paragraph for email
        | :param cc_list: `list` List of emails to CC for this
        | :return: `void`
        """
        self.send = True
        self.subject = subject
        self.description = description
        if cc_list:
            self.cc_list = cc_list
        self.flush()


class PropertyFileManager:
    """
    | Manages server.properties files.
    """
    def __init__(self, filename, properties):
        """
        | Finds and replaces properties lines in server.properties
        |
        | :param filename: File path to server.properties
        | :param properties: `dict` The properties to update
        """
        self.logger = logging.getLogger('mine-init')
        self.filename = filename
        self.properties = properties
        self.lines = ''

    def read(self):
        """
        | Loads the server.properties file into memory.
        """
        print('Opening server properties file at %s' % self.filename)
        try:
            with open(self.filename, 'r') as file:
                self.lines = file.readlines()
        except FileNotFoundError:
            self.logger.error('Error: Server properties file (%s) does not exist.', self.filename)
        except PermissionError:
            self.logger.error('Error: We were not able to access %s, check file permissions or '
                              'escalate your privileges as necessary.', self.filename)

    def set_property(self, name, value):
        """
        | Updates a property in the file contents if it has changed.
        | :param name:
        | :param value:
        """
        rex = re.compile(r'^%s\s*=' % name)
        changed = False
        for index in range(len(self.lines)):
            line = self.lines[index]
            if rex.match(line):
                self.lines[index] = '%s = %s\n' % (name, value)
                self.logger.info('Server property %s updated to %s', name, value)
                changed = True
        if not changed:
            self.lines.append('%s = %s\n' % (name, value))

    def update_properties(self):
        """
        | Checks properties dictionary and feeds them to set_property()
        | :return: `boolean` True or false based on success.
        """

        for prop, value in self.properties:
            self.set_property(prop, value)

    def write(self):
        """
        | Saves the updated file contents to server.properties
        | :return: `void`
        """
        print('Saving changes to %s' % self.filename)
        try:
            with open(self.filename, 'w') as file:
                file.write(''.join(self.lines))
        except FileNotFoundError:
            self.logger.error('Error: Server properties file (%s) does not exist.', self.filename)
        except PermissionError:
            self.logger.error('Error: We were not able to access %s, check file permissions '
                              'or escalate your privileges as necessary.', self.filename)


class MOTDGenerator:
    """
    | An MOTD Generator.
    """
    def __init__(self):
        if os.environ.get('MOTD'):
            self.motd = str(os.environ.get('MOTD'))

    def generate(self):
        """
        | Check for the MOTD property and return, else auto-generate and return.
        | :return: `string` Message of the day.
        """

        if self.motd:
            return self.motd

        game_type = os.environ.get('PACK_NAME', os.environ.get('GAME_TYPE'.lower(), 'modded'))
        version = os.environ.get('VERSION')
        creator = os.environ.get('PACK_CREATOR', 'Routh.IO')
        host = os.environ.get('GAME_HOST', 'Routh.IO')

        return 'A %s %s Minecraft server by %s. Hosted by %s' % game_type, version, creator, host


class PackFileUpdater:
    """
    | Updates files in from a pack distribution to a server volume.
    """
    def __init__(self, source, dest):
        self.source = source
        self.dest = dest
        self.logger = logging.getLogger('mine-init')

    def sync(self):
        """
        | Sync the /dist/pack/server directory with the /server directory.
        | :return: `boolean` This always returns true, for now.
        """
        try:
            result = subprocess.run(
                'rm -rf '
                'mods '
                'scripts '
                'libraries '
                'forge*.jar '
                'minecraft*.jar',
                cwd=self.dest
            )
            result.check_returncode()
        except FileNotFoundError:
            self.logger.error('Error accessing %s directory', self.dest)

        try:
            result = subprocess.run(
                'rsync -av --ignore-existing %s/server/ %s'
                % (self.source, self.dest)
            )
            result.check_returncode()
        except FileNotFoundError:
            self.logger.error('Error accessing %s directory', self.dest)

        try:
            result = subprocess.run(
                'rsync -av %s/luckperms/ %s/luckperms/'
                % (self.source, self.dest)
            )
            result.check_returncode()
        except FileNotFoundError:
            self.logger.error(
                'Error accessing %s/luckperms directory',
                self.dest
            )

        try:
            result = subprocess.run(
                'rsync -av %s/config/ %s/config/'
                % (self.source, self.dest)
            )
            result.check_returncode()
        except FileNotFoundError:
            self.logger.error(
                'Error accessing %s/config directory',
                self.dest
            )

        return True


class PackManager:
    """
    | Manages a Packmaker installation.
    """
    def __init__(self, build_dir, cache_dir='/tmp/packmaker'):
        self.build_dir = build_dir
        self.cache_dir = cache_dir
        self.logger = logging.getLogger('mine-init')

    def install_pack(self):
        """
        | Spawns a subprocess to preinstall the Packmaker pack.
        |
        | :return: `boolean` True or False based on success.
        """
        try:
            result = subprocess.run(
                'packmaker build-server '
                '--cache-dir %s --build-dir %s' % (self.cache_dir, self.build_dir),
                shell=True,
                cwd=self.build_dir
            )
            result.check_returncode()
        except subprocess.CalledProcessError as pm_error:
            self.logger.error(
                'Error with Packmaker installation. See stack trace: \n\n %s',
                pm_error
            )
            return False
        return True



"""
| Main Classes for mine-init
"""
import argparse
import os
import sys
import datetime
import getpass
import socket
import logging
from .utils import PropertyFileManager, MOTDGenerator, PackFileUpdater, PackManager, \
    BufferingSMTPHandler, MakeFileHandler


DIST = os.getenv('DIST', '/dist')
SERVER_DIR = os.getenv('SERVER_DIR', '/server')
PROP_FILE = os.getenv('SERVER_PROPERTIES', '/server/server.properties')
SERVER_PROPS = {
    'motd': MOTDGenerator.generate,
    'server-port': os.getenv('SERVER_PORT', '25565'),
    'allow-nether': os.getenv('ALLOW_NETHER', 'true'),
    'announce-player-achievements': os.getenv('ANNOUNCE_PLAYER_ACHIEVEMENTS', 'true'),
    'enable-command-block': os.getenv('ENABLE_COMMAND_BLOCK', 'false'),
    'spawn-animals': os.getenv('SPAWN_ANIMALS', 'true'),
    'spawn-monsters': os.getenv('SPAWN_MONSTERS', 'true'),
    'spawn-npcs': os.getenv('SPAWN_NPCS', 'true'),
    'generate-structures': os.getenv('GENERATE_STRUCTURES', 'true'),
    'view-distance': os.getenv('VIEW_DISTANCE', '10'),
    'hardcore': os.getenv('HARDCORE', 'false'),
    'max-build-height': os.getenv('MAX_BUILD_HEIGHT', '256'),
    'force-gamemode': os.getenv('FORCE_GAMEMODE', 'false'),
    'hardmax-tick-timecore': os.getenv('MAX_TICK_TIME', '-1'),
    'enable-query': os.getenv('ENABLE_QUERY', 'false'),
    'query.port': os.getenv('QUERY_PORT', '25565'),
    'enable-rcon': os.getenv('ENABLE_RCON', 'true'),
    'rcon.password': os.getenv('RCON_PASSWORD', 'minecraft'),
    'rcon.port': os.getenv('RCON_PORT', '25575'),
    'max-players': os.getenv('MAX_PLAYERS', '40'),
    'max-world-size': os.getenv('MAX_WORLD_SIZE', '5000'),
    'level-name': os.getenv('LEVEL', 'world'),
    'level-seed': os.getenv('SEED', ''),
    'pvp': os.getenv('PVP', 'true'),
    'generator-settings': os.getenv('GENERATOR_SETTINGS', ''),
    'online-mode': os.getenv('ONLINE_MODE', 'true'),
    'allow-flight': os.getenv('ALLOW_FLIGHT', 'true'),
    'level-type': os.getenv('LEVEL_TYPE'.upper(), 'DEFAULT'),
    'white-list': os.getenv('WHITELIST', 'false'),
    'whitelist': os.getenv('WHITELIST', 'false'),
    'spawn-protection': os.getenv('SPAWN_PROTECTION', '0'),
    'difficulty': os.getenv('DIFFICULTY', '1'),
    'gamemode': os.getenv('GAMEMODE', '0')
}


class MineInit:
    """
    | The main class for the mine-init executable.
    """

    def __init__(self):
        self.hostname = socket.getfqdn()
        self.parser = self.build_parser()
        self.ppid = os.getppid()
        args = self.get_args()
        try:
            func = args.func
        except AttributeError:
            self.parser.error('Too few arguments, see "mine-init -h"')
        self.logger = self.setup_logging(args)
        func(args)

    def setup_logging(self, args):
        """
        | Setup logging
        | :return: logging.Logger instance
        """
        log_dir = args.log_dir
        log_filename = datetime.datetime.now().strftime('%Y-%m-%d') + '.log'
        logger = logging.getLogger('mine-init')
        logger.setLevel(args.log_level)
        log_format = logging.Formatter('%(asctime)s : %(levelname)s : %(message)s')

        file_handler = MakeFileHandler(filename=os.path.join(log_dir, log_filename))
        file_handler.setLevel(args.log_level)
        file_handler.setFormatter(log_format)
        logger.addHandler(file_handler)

        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setLevel(args.log_level)
        stream_handler.setFormatter(log_format)
        logger.addHandler(stream_handler)

        self.smtp_handler = BufferingSMTPHandler(
            args.mail_server,
            args.mail_port,
            args.src_email,
            args.dest_email,
            1000
        )
        self.smtp_handler.setLevel(args.log_level)
        self.smtp_handler.setFormatter(log_format)
        logger.addHandler(self.smtp_handler)

        return logger

    def get_args(self):
        """
        | Return parsed arguments object.
        | :return: `object` argparser parsed arguments.
        """
        return self.parser.parse_args()

    def build_parser(self):
        """
        | Build the main argparser for the application and return as an object.
        | :return: `object` argparser
        """
        parser = argparse.ArgumentParser(prog='mine-init')
        subparsers = parser.add_subparsers(dest='subcommands')
        parser.add_argument(
            '-v', '--verbose',
            dest='verbose',
            action='store_true',
            help='Make things chatty.'
        )

        parser.add_argument(
            '--dest-email',
            dest='dest_email',
            type=str,
            required=True,
            help='The email address to send the backup report to.'
        )

        parser.add_argument(
            '--source-email',
            dest='src_email',
            type=str,
            default="%s@%s" % (getpass.getuser(), socket.getfqdn()),
            help='(optional) The email address to send the backup report from. Defaults to '
                 '%s@%s' % (getpass.getuser(), socket.getfqdn())
        )

        parser.add_argument(
            '--mail-server',
            dest='mail_server',
            type=str,
            default='localhost',
            help='(optional) The mail server to use to send mail. Defaults to localhost.'
        )

        parser.add_argument(
            '--mail-port',
            dest='mail_port',
            type=int,
            default=25,
            help='(optional) The port on which to connect to the mail server. Defaults to 25.'
        )

        parser.add_argument(
            '--log-level',
            dest='log_level',
            type=str,
            default='INFO',
            help='(optional) The logging level, defaults to INFO.'
        )

        parser.add_argument(
            '--log-dir',
            dest='log_dir',
            type=str,
            default='/db_backups/logs',
            help='(optional) The directory to place logs. Defaults to /db_backups/logs'
        )

        parser_start = subparsers.add_parser(
            'start',
            help='Starts and monitors a Minecraft instance.'
        )
        parser_start.add_argument(
            '-x', '--no-update',
            action='store_true',
            dest='update',
            help='Disable update of the server directory from the distribution directory.'
                 ' Defaults to false.'
        )
        parser_start.add_argument(
            '-d', '--dist', '--distribution-dir',
            type=str,
            dest='dist',
            default='/dist',
            help='The directory containing the distribution files, like default server.properties,'
                 ' configs, and as infrequently as possible (to prevent bloat and legal issues)'
                 ' mods. Defaults to "/dist".'
        )
        parser_start.add_argument(
            '-s', '--server', '--server-dir',
            type=str,
            dest='server',
            default='/server',
            help='The directory where the server root will be (or is). Defaults to "/server".'
        )
        parser_start.set_defaults(func=self.start)
        return parser

    @staticmethod
    def start(args):
        """
        | Starts and updates a Minecraft instance.
        | :param args: `object` argparse arguments object.
        | :return:
        """
        source_dir = DIST
        if args.dist:
            source_dir = args.dist

        server_dir = SERVER_DIR
        if args.server:
            server_dir = args.server

        pack_manager = PackManager(source_dir)
        pack_manager.install_pack()

        pack_update = PackFileUpdater(source_dir, server_dir)
        pack_update.sync()

        prop_manager = PropertyFileManager(PROP_FILE, SERVER_PROPS)
        prop_manager.read()
        prop_manager.update_properties()
        prop_manager.write()



"""
| Minecraft server configuration, update synchronisation and startup logic
|
| Maintained by Routh.IO
"""
from .core import MineInit


def main():
    """
    | Script entrypoint
    """
    return MineInit()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['mine_init']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['mine-init = mine_init:main']}

setup_kwargs = {
    'name': 'mine-init',
    'version': '0.1.5',
    'description': 'A docker friendly startup routine for Minecraft servers.',
    'long_description': '===========\n mine-init\n===========\n----------------------------------------------------------\n A docker friendly startup routine for Minecraft servers.\n----------------------------------------------------------\n\n|build-status| |coverage|\n\nMain Documentation\n==================\n\n`Main Index`_\n\n.. |build-status| image:: https://gitlab.routh.io/minecraft/tools/mine-init/badges/master/pipeline.svg\n    :target: https://gitlab.routh.io/minecraft/tools/mine-init/pipelines\n\n.. |coverage| image:: https://gitlab.routh.io/minecraft/tools/mine-init/badges/master/coverage.svg\n    :target: http://minecraft.pages.routh.io/tools/mine-init/reports/\n    :alt: Coverage status\n\n.. _Main Index: http://minecraft.pages.routh.io/tools/mine-init/\n',
    'author': 'Chris Routh',
    'author_email': 'chris@routh.io',
    'url': 'https://gitlab.routh.io/minecraft/tools/mine-init',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
